import { FormBuilder, FormControl } from '@angular/forms';
import { reduce } from 'lodash/fp';
import * as fmdModels from './../models';

/*
 * FORM UTILITIES
 *
 * Exports
 * -------
 * buildFormGroup(metadata)  - builds FormGroups from metdata, recursively if necessary
 *
 * Variable names
 * --------------
 * metaF   = metadata for Field
 * metaG   = metadata for Group (possibly nested)
 * metaFoG = metadata for Field Or Group
 *
 */

const fb = new FormBuilder();

// ---------------------------------------------------------------------------------------------------------------------
// Helper Funstions

const addProp = (obj, key, val) => { obj[key] = val; return obj; };

// Is Group
// Helper function to distinguish group from field
const isGroup = metaFoG => metaFoG.meta;

// Add Missing Names
// Helper function to add any missing 'name' properties to Fields and Groups using property's key, recursively
const addNameIfMissing = (metaFoG, key) => metaFoG.name ? metaFoG : addProp(metaFoG, 'name', key);
const addNameToSelfAndChildren = ( [key, metaFoG] ) => {
	metaFoG = addNameIfMissing(metaFoG, key);
	if (isGroup(metaFoG)) {
		metaFoG.meta = addMissingNames(metaFoG.meta); // Recursion
	}
	return [key, metaFoG];
};
const addMissingNames = metaG => Object.entries(metaG)
	.map(addNameToSelfAndChildren)
	.reduce((res, [key, val]) => addProp(res, key, val), {});

// ---------------------------------------------------------------------------------------------------------------------
// Raw Model (or Mapped Raw Model) to Automatic Metadata

const isScalar = val => typeof val === 'boolean' || typeof val === 'number' || typeof val === 'string';

const keyValPairToMeta = (val, key) => ({ name: key, [isScalar(val) ? 'value' : 'meta']: val });
const keyValPairToMetaRecursive = ( [key, val] ) => {
	const innerVal = isScalar(val) ? val : rawModelAutoMeta(val);
	const metaVal = keyValPairToMeta(innerVal, key);
	return [key, metaVal];
};
const rawModelAutoMeta = model => Object.entries(model)
	.map(keyValPairToMetaRecursive)
	.reduce((res, [key, val]) => addProp(res, key, val), {});


// If value is complex, should we make name & meta, not name & value,
// as form fields can't generally express complex values

// ---------------------------------------------------------------------------------------------------------------------
// Combine automatically generated metadata with overrides

const combineMetaForField = (metaF, metaFextra) => Object.assign(metaF, metaFextra);
const combineExtraMeta = (metaG, metaExtra) => {
	let combinedMeta = {};
	Object.entries(metaExtra).forEach(([key, val]) => {
		if (typeof metaG[key] === 'object') {
			combinedMeta[key] = val.meta ?
				combineExtraMeta(metaG[key].meta, val.meta) :
				combineMetaForField(metaG[key], val);
		}
	});
	return {...metaG, ...combinedMeta};
};

// ---------------------------------------------------------------------------------------------------------------------
// Form Metadata Factory

const buildModeledField = metaF => {
	// TODO: Add validity check
	const modeledMetaF = new fmdModels[metaF.type](metaF);
	return modeledMetaF;
};

// Build Form Group Member
const buildModeledGroupMember = metaFoG => isGroup(metaFoG) ? _buildModeledGroup(metaFoG.meta) : buildModeledField(metaFoG);

// Build Form Group
const buildModeledGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildFormGroupMember(metaFoG) });
const _buildModeledGroup = metaG => reduce(buildModeledGroupReducerIteree, {}, metaG);
const buildModeledGroup = metaG => _buildModeledGroup(addMissingNames(metaG));

// ---------------------------------------------------------------------------------------------------------------------
// Functions which build FormControls and FormGroups

// Build Form Control
// TODO: Flesh out function to build validators
const buildControlState = metaF => ({ value: metaF.value || metaF.default, disabled: metaF.isDisabled });
const buildValidators = metaF => ({
	validators: null,
	asyncValidators: null,
	updateOn: 'blur'
});
const buildFormControl = metaF => new FormControl(buildControlState(metaF) /*, buildValidators(metaF) */);

// Build Form Group Member
const buildFormGroupMember = metaFoG => isGroup(metaFoG) ? _buildFormGroup(metaFoG.meta) : buildFormControl(metaFoG);

// Build Form Group
const buildFormGroupReducerIteree = (res, metaFoG) => Object.assign(res, { [metaFoG.name]: buildFormGroupMember(metaFoG) });
const _buildFormGroup = metaG => fb.group(reduce(buildFormGroupReducerIteree, {}, metaG));
const buildFormGroup = metaG => _buildFormGroup(addMissingNames(metaG));

// ---------------------------------------------------------------------------------------------------------------------
// Exports

export { buildFormGroup };
